from .missing_element_base import MissingElementBase
from .missing_fieldset_entry import MissingFieldsetEntry

class MissingFieldset(MissingElementBase):
    def __repr__(self):
        if self._key:
            return f"<class MissingFieldset key={self._key}>"

        return '<class MissingFieldset>'

    def entries(self, _key=None):
        return []

    def entry(self, key=None):
        return MissingFieldsetEntry(key, self)

    def optional_entry(self, _key=None):
        return None

    def required_entry(self, _key=None):
        self._parent._missing_error(self)



from ..constants import (
    DOCUMENT,
    FIELD,
    FIELDSET,
    FIELDSET_ENTRY,
    LIST,
    LIST_ITEM,
    MULTILINE_FIELD_BEGIN,
    SECTION
)

DISPLAY = 'Display Line'
EMPHASIZE = 'Emphasize Line'
INDICATE = 'Indicate Line'
OMISSION = 'Insert Omission'
QUESTION = 'Question Line'

class Reporter:
    def __init__(self, context):
        self._context = context
        self._index = [None] * self._context.line_count
        self._snippet = [None] * self._context.line_count

        self._build_index()

    def _build_index(self):
        # TODO: Put this into the root class scope?
        def index_comments(element):
            if 'comments' in element:
                for comment in element['comments']:
                    self._index[comment['line']] = comment

        # TODO: Put this into the root class scope?
        def traverse(section):
            for element in section['elements']:
                index_comments(element)

                self._index[element['line']] = element

                if element['type'] == SECTION:
                    traverse(element)
                elif element['type'] == FIELD:
                    if 'continuations' in element:
                        for continuation in element['continuations']:
                            self._index[continuation['line']] = continuation
                elif element['type'] == MULTILINE_FIELD_BEGIN:
                    # Missing when reporting an unterminated multiline field
                    if 'end' in element:
                        self._index[element['end']['line']] = element['end']

                    if 'lines' in element:
                        for line in element['lines']:
                            self._index[line['line']] = line

                elif element['type'] == LIST:
                    if 'items' in element:
                        for item in element['items']:
                            index_comments(item)

                            self._index[item['line']] = item

                            if 'continuations' in item:
                                for continuation in item['continuations']:
                                    self._index[continuation['line']] = continuation

                elif element['type'] == FIELDSET:
                    if 'entries' in element:
                        for entry in element['entries']:
                            index_comments(entry)

                            self._index[entry['line']] = entry

                            if 'continuations' in entry:
                                for continuation in entry['continuations']:
                                    self._index[continuation['line']] = continuation

        traverse(self._context.document)

        for meta in self._context.meta:
            self._index[meta['line']] = meta

    def _tag_continuations(self, element, tag):
        scan_line = element['line'] + 1

        if not 'continuations' in element:
            return scan_line

        for continuation in element['continuations']:
            while scan_line < continuation['line']:
                self._snippet[scan_line] = tag
                scan_line += 1

            self._snippet[continuation['line']] = tag
            scan_line += 1

        return scan_line

    def _tag_continuables(self, element, collection, tag):
        scan_line = element['line'] + 1

        if not collection in element:
            return scan_line

        for continuable in element[collection]:
            while scan_line < continuable['line']:
                self._snippet[scan_line] = tag
                scan_line += 1

            self._snippet[continuable['line']] = tag

            scan_line = self._tag_continuations(continuable, tag)

        return scan_line

    def _tag_children(self, element, tag):
        if element['type'] == FIELD or element['type'] == LIST_ITEM or element['type'] == FIELDSET_ENTRY:
            return self._tag_continuations(element, tag)
        elif element['type'] == LIST:
            return self._tag_continuables(element, 'items', tag)
        elif element['type'] == FIELDSET and 'entries' in element:
            return self._tag_continuables(element, 'entries', tag)
        elif element['type'] == MULTILINE_FIELD_BEGIN:
            if 'lines' in element:
                for line in element['lines']:
                    self._snippet[line['line']] = tag

            if 'end' in element:
                self._snippet[element['end']['line']] = tag
                return element['end']['line'] + 1
            elif 'lines' in element:
                return element['lines'][-1]['line'] + 1
            else:
                return element['line'] + 1
        elif element['type'] == SECTION:
            return self._tag_section(element, tag)

    def _tag_section(self, section, tag, recursive=True):
        scan_line = section['line'] + 1

        for element in section['elements']:
            while scan_line < element['line']:
                self._snippet[scan_line] = tag
                scan_line += 1

            if not recursive and element['type'] == SECTION:
                break

            self._snippet[element['line']] = tag

            scan_line = self._tag_children(element, tag)

        return scan_line

    def indicate_line(self, element):
        self._snippet[element['line']] = INDICATE
        return self

    def question_line(self, element):
        self._snippet[element['line']] = QUESTION
        return self

    def report_comments(self, element):
        self._snippet[element['line']] = INDICATE
        for comment in element['comments']:
            self._snippet[comment['line']] = EMPHASIZE

        return self

    def report_element(self, element):
        self._snippet[element['line']] = EMPHASIZE
        self._tag_children(element, INDICATE)

        return self

    def report_elements(self, elements):
        for element in elements:
            self._snippet[element['line']] = EMPHASIZE
            self._tag_children(element, INDICATE)

        return self

    def report_line(self, instruction):
        self._snippet[instruction['line']] = EMPHASIZE

        return self

    def report_multiline_value(self, element):
        for line in element['lines']:
            self._snippet[line['line']] = EMPHASIZE

        return self

    def report_missing_element(self, parent):
        if parent['type'] != DOCUMENT:
            self._snippet[parent['line']] = INDICATE

        if parent['type'] == SECTION:
            self._tag_section(parent, QUESTION, False)
        else:
            self._tag_children(parent, QUESTION)

        return self

    def snippet(self):
        if all(tag is None for tag in self._snippet):
            for line in range(len(self._snippet)):
                self._snippet[line] = QUESTION
        else:
            # TODO: Possibly better algorithm for this

            for line, tag in enumerate(self._snippet):
                if tag is not None:
                    continue

                if(line + 2 < self._context.line_count and self._snippet[line + 2] is not None and self._snippet[line + 2] != DISPLAY or
                   line - 2 > 0 and self._snippet[line - 2] is not None and self._snippet[line - 2] != DISPLAY or
                   line + 1 < self._context.line_count and self._snippet[line + 1] is not None and self._snippet[line + 1] != DISPLAY or
                   line - 1 > 0 and self._snippet[line - 1] is not None and self._snippet[line - 1] != DISPLAY):
                    self._snippet[line] = DISPLAY
                elif line + 3 < self._context.line_count and self._snippet[line + 3] is not None and self._snippet[line + 3] != DISPLAY:
                    self._snippet[line] = OMISSION

            if self._snippet[-1] is None:
                self._snippet[-1] = OMISSION

        return self._print()



# Added to 0-indexed indices in a few places
HUMAN_INDEXING = 1

# Selection indices
BEGIN = 0
END = 1

# Instruction types
COMMENT = 'Comment'
CONTINUATION = 'Continuation'
DOCUMENT = 'Document'
EMPTY_ELEMENT = 'Empty Element'
FIELD = 'Field'
FIELDSET = 'Fieldset'
FIELDSET_ENTRY = 'Fieldset Entry'
LIST = 'List'
LIST_ITEM = 'List Item'
MULTILINE_FIELD_BEGIN = 'Multiline Field Begin'
MULTILINE_FIELD_END = 'Multiline Field End'
MULTILINE_FIELD_VALUE = 'Multiline Field Value'
SECTION = 'Section'
UNPARSED = 'Unparsed'

PRETTY_TYPES = {
    DOCUMENT: 'document',
    EMPTY_ELEMENT: 'empty_element',
    FIELD: 'field',
    FIELDSET: 'fieldset',
    FIELDSET_ENTRY: 'fieldset_entry',
    LIST: 'list',
    LIST_ITEM: 'list_item',
    MULTILINE_FIELD_BEGIN: 'field',
    SECTION: 'section'
}



from html import escape

from ..constants import BEGIN, END, HUMAN_INDEXING
from .reporter import EMPHASIZE, INDICATE, OMISSION, QUESTION, Reporter


class HtmlReporter(Reporter):
    def _line(self, line, tag):
        if tag == OMISSION:
            return self._markup('...', '...')

        number = str(line + HUMAN_INDEXING)
        instruction = self._index[line]

        content = ''
        if instruction:
            content = self._context.input[instruction['ranges']['line'][BEGIN]:instruction['ranges']['line'][END]]

        if tag == EMPHASIZE:
            tag_class = 'eno-report-line-emphasized'
        elif tag == INDICATE:
            tag_class = 'eno-report-line-indicated'
        elif tag == QUESTION:
            tag_class = 'eno-report-line-questioned'
        else:
            tag_class = ''

        return self._markup(number, content, tag_class)

    def _markup(self, gutter, content, tag_class=''):
        return (
            f"<div class=\"eno-report-line {tag_class}\">"
            f"<div class=\"eno-report-gutter\">{gutter.rjust(10)}</div>"
            f"<div class=\"eno-report-content\">{escape(content)}</div>"
            '</div>'
        )

    def _print(self):
        columns_header = self._markup(self._context.messages.gutter_header, self._context.messages.content_header)
        snippet = '\n'.join(self._line(line, tag) for line, tag in enumerate(self._snippet) if tag is not None)

        if self._context.source:
            return f"<div><div>{self._context.source}</div><pre class=\"eno-report\">{columns_header}{snippet}</pre></div>"

        return f"<pre class=\"eno-report\">{columns_header}{snippet}</pre>"



#  GENERATED ON 2019-05-29T07:40:09 - DO NOT EDIT MANUALLY

content_header = 'Content'
expected_document = 'The document was expected.'
expected_empty = 'An empty element was expected.'
expected_field = 'A field was expected.'
expected_fields = 'Only fields were expected.'
expected_fieldset = 'A fieldset was expected.'
expected_fieldset_entry = 'A fieldset entry was expected.'
expected_fieldsets = 'Only fieldsets were expected.'
expected_list = 'A list was expected.'
expected_list_item = 'A list item was expected.'
expected_lists = 'Only lists were expected.'
expected_section = 'A section was expected.'
expected_sections = 'Only sections were expected.'
expected_single_element = 'Only a single element was expected.'
expected_single_empty = 'Only a single empty element was expected.'
expected_single_field = 'Only a single field was expected.'
expected_single_fieldset = 'Only a single fieldset was expected.'
expected_single_fieldset_entry = 'Only a single fieldset entry was expected.'
expected_single_list = 'Only a single list was expected.'
expected_single_section = 'Only a single section was expected.'
gutter_header = 'Line'
missing_comment = 'A required comment for this element is missing.'
missing_element = 'A single element is required - it can have any key.'
missing_empty = 'A single empty element is required - it can have any key.'
missing_field = 'A single field is required - it can have any key.'
missing_fieldset = 'A single fieldset is required - it can have any key.'
missing_fieldset_entry = 'A single fieldset entry is required - it can have any key.'
missing_list = 'A single list is required - it can have any key.'
missing_section = 'A single section is required - it can have any key.'
unexpected_element = 'This element was not expected, make sure it is at the right place in the document and that its key is not mis-typed.'
comment_error = lambda message: f"There is a problem with the comment of this element: {message}"
cyclic_dependency = lambda line, key: f"In line {line} '{key}' is copied into itself."
expected_empty_with_key = lambda key: f"An empty element with the key '{key}' was expected."
expected_field_with_key = lambda key: f"A field with the key '{key}' was expected."
expected_fields_with_key = lambda key: f"Only fields with the key '{key}' were expected."
expected_fieldset_with_key = lambda key: f"A fieldset with the key '{key}' was expected."
expected_fieldsets_with_key = lambda key: f"Only fieldsets with the key '{key}' were expected."
expected_list_with_key = lambda key: f"A list with the key '{key}' was expected."
expected_lists_with_key = lambda key: f"Only lists with the key '{key}' were expected."
expected_section_with_key = lambda key: f"A section with the key '{key}' was expected."
expected_sections_with_key = lambda key: f"Only sections with the key '{key}' were expected."
expected_single_element_with_key = lambda key: f"Only a single element with the key '{key}' was expected."
expected_single_empty_with_key = lambda key: f"Only a single empty element with the key '{key}' was expected."
expected_single_field_with_key = lambda key: f"Only a single field with the key '{key}' was expected."
expected_single_fieldset_entry_with_key = lambda key: f"Only a single fieldset entry with the key '{key}' was expected."
expected_single_fieldset_with_key = lambda key: f"Only a single fieldset with the key '{key}' was expected."
expected_single_list_with_key = lambda key: f"Only a single list with the key '{key}' was expected."
expected_single_section_with_key = lambda key: f"Only a single section with the key '{key}' was expected."
invalid_line = lambda line: f"Line {line} does not follow any specified pattern."
key_error = lambda message: f"There is a problem with the key of this element: {message}"
missing_element_for_continuation = lambda line: f"Line {line} contains a line continuation without a continuable element being specified before."
missing_element_with_key = lambda key: f"The element '{key}' is missing - in case it has been specified look for typos and also check for correct capitalization."
missing_empty_with_key = lambda key: f"The empty element '{key}' is missing - in case it has been specified look for typos and also check for correct capitalization."
missing_field_value = lambda key: f"The field '{key}' must contain a value."
missing_field_with_key = lambda key: f"The field '{key}' is missing - in case it has been specified look for typos and also check for correct capitalization."
missing_fieldset_entry_value = lambda key: f"The fieldset entry '{key}' must contain a value."
missing_fieldset_entry_with_key = lambda key: f"The fieldset entry '{key}' is missing - in case it has been specified look for typos and also check for correct capitalization."
missing_fieldset_for_fieldset_entry = lambda line: f"Line {line} contains a fieldset entry without a fieldset being specified before."
missing_fieldset_with_key = lambda key: f"The fieldset '{key}' is missing - in case it has been specified look for typos and also check for correct capitalization."
missing_list_for_list_item = lambda line: f"Line {line} contains a list item without a list being specified before."
missing_list_item_value = lambda key: f"The list '{key}' may not contain empty items."
missing_list_with_key = lambda key: f"The list '{key}' is missing - in case it has been specified look for typos and also check for correct capitalization."
missing_section_with_key = lambda key: f"The section '{key}' is missing - in case it has been specified look for typos and also check for correct capitalization."
non_section_element_not_found = lambda line, key: f"In line {line} the non-section element '{key}' should be copied, but it was not found."
section_hierarchy_layer_skip = lambda line: f"Line {line} starts a section that is more than one level deeper than the current one."
section_not_found = lambda line, key: f"In line {line} the section '{key}' should be copied, but it was not found."
two_or_more_templates_found = lambda key: f"There are at least two elements with the key '{key}' that qualify for being copied here, it is not clear which to copy."
unterminated_escaped_key = lambda line: f"In line {line} the key of an element is escaped, but the escape sequence is not terminated until the end of the line."
unterminated_multiline_field = lambda key, line: f"The multiline field '{key}' starting in line {line} is not terminated until the end of the document."
value_error = lambda message: f"There is a problem with the value of this element: {message}"


from .elements.element_base import ElementBase
from .elements.list import List
from .elements.value_element_base import ValueElementBase
from .elements.missing.missing_element_base import MissingElementBase
from .elements.missing.missing_list import MissingList
from .elements.missing.missing_value_element_base import MissingValueElementBase

def _register(name: str, function):
    if name == 'string':
        raise ValueError("You cannot register 'string' as a type/loader with enolib as this conflicts with the native string type accessors.")

    setattr(ElementBase, f"{name}_key", lambda self: self.key(function))
    setattr(ElementBase, f"optional_{name}_comment",  lambda self: self.optional_comment(function))
    setattr(ElementBase, f"required_{name}_comment",  lambda self: self.required_comment(function))
    setattr(ValueElementBase, f"optional_{name}_value", lambda self: self.optional_value(function))
    setattr(ValueElementBase, f"required_{name}_value", lambda self: self.required_value(function))
    setattr(List, f"optional_{name}_values", lambda self: self.optional_values(function))
    setattr(List, f"required_{name}_values", lambda self: self.required_values(function))
    setattr(MissingElementBase, f"{name}_key", MissingElementBase.string_key)
    setattr(MissingElementBase, f"optional_{name}_comment", MissingElementBase.optional_string_comment)
    setattr(MissingElementBase, f"required_{name}_comment", MissingElementBase.required_string_comment)
    setattr(MissingValueElementBase, f"optional_{name}_value", MissingValueElementBase.optional_string_value)
    setattr(MissingValueElementBase, f"required_{name}_value", MissingValueElementBase.required_string_value)
    setattr(MissingList, f"optional_{name}_values", MissingList.optional_string_values)
    setattr(MissingList, f"required_{name}_values", MissingList.required_string_values)

# TODO: Specs for different register signatures (see below)

def register(*definitions_list, **definitions_dict):
    for definition in definitions_list:
        if callable(definition):
            _register(definition.__name__, definition)
        else:
            for name, function in definition.items():
                _register(name, function)

    for name, function in definitions_dict.items():
        _register(name, function)



class Error(Exception):
    def __init__(self, text, snippet, selection):
        super().__init__(f"{text}\n\n{snippet}")

        self.cursor = selection['from']
        self.message = f"{text}\n\n{snippet}"
        self.selection = selection
        self.snippet = snippet
        self.text = text


class ParseError(Error):
    pass


class ValidationError(Error):
    pass



from ..errors.validation import Validation
from .element_base import ElementBase


class ValueElementBase(ElementBase):
    def _print_value(self):
        value = self._context.value(self._instruction)

        if not value:
            return 'None'

        if len(value) > 14:
            value = f"{value[0:11]}..."

        return value.replace("\n", '\n')

    def _value(self, loader=None, *, required):
        self._touched = True

        value = self._context.value(self._instruction)

        if value is None:
            if required:
                raise Validation.missing_value(self._context, self._instruction)
            else:
                return None

        # TODO: Check if ValueError is 1) a valid pythonic approach 2) the best choice

        if not loader:
            return value

        try:
            return loader(value)
        except ValueError as message:
            raise Validation.value_error(self._context, message, self._instruction)

    def optional_string_value(self):
        return self._value(required=False)

    def optional_value(self, loader):
        return self._value(loader, required=False)

    def required_string_value(self):
        return self._value(required=True)

    def required_value(self, loader):
        return self._value(loader, required=True)

    def value_error(self, message):
        if callable(message):
            message = message(self._context.value(self._instruction))

        return Validation.value_error(self._context, message, self._instruction)



from .missing_element_base import MissingElementBase

class MissingEmpty(MissingElementBase):
    def __repr__(self):
        if self._key:
            return f"<class MissingEmpty key={self._key}>"

        return '<class MissingEmpty>'



from .missing_element_base import MissingElementBase
from . import missing_empty
from . import missing_field
from . import missing_fieldset
from . import missing_list
from . import missing_section

class MissingSectionElement(MissingElementBase):
    def __repr__(self):
        if self._key:
            return f"<class MissingSectionElement key={self._key}>"

        return '<class MissingSectionElement>'

    def to_empty(self):
        return missing_empty.MissingEmpty(self._key, self._parent)

    def to_field(self):
        return missing_field.MissingField(self._key, self._parent)

    def to_fieldset(self):
        return missing_fieldset.MissingFieldset(self._key, self._parent)

    def to_list(self):
        return missing_list.MissingList(self._key, self._parent)

    def to_section(self):
        return missing_section.MissingSection(self._key, self._parent)

    def yields_empty(self):
        return True # TODO: here and below - raise instead?

    def yields_field(self):
        return True

    def yields_fieldset(self):
        return True

    def yields_list(self):
        return True

    def yields_section(self):
        return True



from .missing_element_base import MissingElementBase
from . import missing_empty
from . import missing_field
from . import missing_fieldset
from . import missing_list
from . import missing_section_element

class MissingSection(MissingElementBase):
    def __repr__(self):
        if self._key:
            return f"<class MissingSection key={self._key}>"

        return '<class MissingSection>'

    def element(self, key=None):
        return missing_section_element.MissingSectionElement(key, self)

    def elements(self, _key=None):
        return []

    def empty(self, key=None):
        return missing_empty.MissingEmpty(key, self)

    def field(self, key=None):
        return missing_field.MissingField(key, self)

    def fields(self, _key=None):
        return []

    def fieldset(self, key=None):
        return missing_fieldset.MissingFieldset(key, self)

    def fieldsets(self, _key=None):
        return []

    def list(self, key=None):
        return missing_list.MissingList(key, self)

    def lists(self, _key=None):
        return []

    def optional_element(self, _key=None):
        return None

    def optional_empty(self, _key=None):
        return None

    def optional_field(self, _key=None):
        return None

    def optional_fieldset(self, _key=None):
        return None

    def optional_list(self, _key=None):
        return None

    def optional_section(self, _key=None):
        return None

    def required_element(self, _key=None):
        self._parent._missing_error(self)

    def required_empty(self, _key=None):
        self._parent._missing_error(self)

    def required_field(self, _key=None):
        self._parent._missing_error(self)

    def required_fieldset(self, _key=None):
        self._parent._missing_error(self)

    def required_list(self, _key=None):
        self._parent._missing_error(self)

    def required_section(self, _key=None):
        self._parent._missing_error(self)

    def section(self, key=None):
        return MissingSection(key, self)

    def sections(self, _key=None):
        return []



from .context import Context
from .elements.element import Element
from .constants import (
    BEGIN,
    END,
    FIELD,
    FIELDSET,
    LIST,
    MULTILINE_FIELD_BEGIN,
    SECTION
)

def check_multiline_field_by_line(field, line):
    if line < field['line'] or line > field['end']['line']:
        return False

    if line == field['line']:
        return { 'element': field, 'instruction': field }

    if line == field['end']['line']:
        return { 'element': field, 'instruction': field['end'] }

    return {
        'element': field,
        'instruction': next((l for l in field['lines'] if l['line'] == line), None)
    }

def check_multiline_field_by_index(field, index):
    if index < field['ranges']['line'][BEGIN] or index > field['end']['ranges']['line'][END]:
        return False

    if index <= field['ranges']['line'][END]:
        return { 'element': field, 'instruction': field }

    if index >= field['end']['ranges']['line'][BEGIN]:
        return { 'element': field, 'instruction': field['end'] }

    return {
        'element': field,
        'instruction': next((l for l in field['lines'] if index <= l['ranges']['line'][END]), None)
    }

def check_field_by_line(field, line):
    if line < field['line']:
        return False

    if line == field['line']:
        return { 'element': field, 'instruction': field }

    if not 'continuations' in field or line > field['continuations'][-1]['line']:
        return False

    for continuation in field['continuations']:
        if line == continuation['line']:
            return { 'element': field, 'instruction': continuation }
        if line < continuation['line']:
            return { 'element': field, 'instruction': None }

def check_field_by_index(field, index):
    if index < field['ranges']['line'][BEGIN]:
        return False

    if index <= field['ranges']['line'][END]:
        return { 'element': field, 'instruction': field }

    if not 'continuations' in field or index > field['continuations'][-1]['ranges']['line'][END]:
        return False

    for continuation in field['continuations']:
        if index < continuation['ranges']['line'][BEGIN]:
            return { 'element': field, 'instruction': None }
        if index <= continuation['ranges']['line'][END]:
            return { 'element': field, 'instruction': continuation }

def check_fieldset_by_line(fieldset, line):
    if line < fieldset['line']:
        return False

    if line == fieldset['line']:
        return { 'element': fieldset, 'instruction': fieldset }

    if not 'entries' in fieldset or line > fieldset['entries'][-1]['line']:
        return False

    for entry in fieldset['entries']:
        if line == entry['line']:
            return { 'element': entry, 'instruction': entry }

        if line < entry['line']:
            if 'comments' in entry and line >= entry['comments'][0]['line']:
                for comment in entry['comments']:
                    if line == comment['line']:
                        return { 'element': entry, 'instruction': comment }

            return { 'element': fieldset, 'instruction': None }

        match_in_entry = check_field_by_line(entry, line)

        if match_in_entry:
            return match_in_entry

def check_fieldset_by_index(fieldset, index):
    if index < fieldset['ranges']['line'][BEGIN]:
        return False

    if index <= fieldset['ranges']['line'][END]:
        return { 'element': fieldset, 'instruction': fieldset }

    if not 'entries' in fieldset or index > fieldset['entries'][-1]['ranges']['line'][END]:
        return False

    for entry in fieldset['entries']:
        if index < entry['ranges']['line'][BEGIN]:
            if 'comments' in entry and index >= entry['comments'][0]['ranges']['line'][BEGIN]:
                for comment in entry['comments']:
                    if index <= comment['ranges']['line'][END]:
                        return { 'element': entry, 'instruction': comment }

            return { 'element': fieldset, 'instruction': None }

        if index <= entry['ranges']['line'][END]:
            return { 'element': entry, 'instruction': entry }

        match_in_entry = check_field_by_index(entry, index)

        if match_in_entry:
            return match_in_entry

def check_list_by_line(list, line):
    if line < list['line']:
        return False

    if line == list['line']:
        return { 'element': list, 'instruction': list }

    if not 'items' in list or line > list['items'][-1]['line']:
        return False

    for item in list['items']:
        if line == item['line']:
            return { 'element': item, 'instruction': item }

        if line < item['line']:
            if 'comments' in item and line >= item['comments'][0]['line']:
                for comment in item['comments']:
                    if line == comment['line']:
                        return { 'element': item, 'instruction': comment }

            return { 'element': list, 'instruction': None }

        match_in_item = check_field_by_line(item, line)

        if match_in_item:
            return match_in_item

def check_list_by_index(list, index):
    if index < list['ranges']['line'][BEGIN]:
        return False

    if index <= list['ranges']['line'][END]:
        return { 'element': list, 'instruction': list }

    if not 'items' in list or index > list['items'][-1]['ranges']['line'][END]:
        return False

    for item in list['items']:
        if index < item['ranges']['line'][BEGIN]:
            if 'comments' in item and index >= item['comments'][0]['ranges']['line'][BEGIN]:
                for comment in item['comments']:
                    if index <= comment['ranges']['line'][END]:
                        return { 'element': item, 'instruction': comment }

            return { 'element': list, 'instruction': None }

        if index <= item['ranges']['line'][END]:
            return { 'element': item, 'instruction': item }

        match_in_item = check_field_by_index(item, index)

        if match_in_item:
            return match_in_item

def check_in_section_by_line(section, line):
    for element in reversed(section['elements']):
        if 'comments' in element:
            if line < element['comments'][0]['line']:
                continue
            if line <= element['comments'][-1]['line']:
                for comment in element['comments']:
                    if line == comment['line']:
                        return { 'element': element, 'instruction': comment }

        if element['line'] > line:
            continue

        if element['line'] == line:
            return { 'element': element, 'instruction': element }

        if element['type'] is FIELD:
            match_in_field = check_field_by_line(element, line)
            if match_in_field:
                return match_in_field
        elif element['type'] is FIELDSET:
            match_in_fieldset = check_fieldset_by_line(element, line)
            if match_in_fieldset:
                return match_in_fieldset
        elif element['type'] is LIST:
            match_in_list = check_list_by_line(element, line)
            if match_in_list:
                return match_in_list
        elif element['type'] is MULTILINE_FIELD_BEGIN:
            if not 'template' in element:  # TODO: More elegant copy detection?
                match_in_multiline_field = check_multiline_field_by_line(element, line)
                if match_in_multiline_field:
                    return match_in_multiline_field
        elif element['type'] is SECTION:
            return check_in_section_by_line(element, line)

        break

    return { 'element': section, 'instruction': None }

def check_in_section_by_index(section, index):
    for element in reversed(section['elements']):
        if 'comments' in element:
            if index < element['comments'][0]['ranges']['line'][BEGIN]:
                continue
            if index <= element['comments'][-1]['ranges']['line'][END]:
                for comment in element['comments']:
                    if index <= comment['ranges']['line'][END]:
                        return { 'element': element, 'instruction': comment }

        if index < element['ranges']['line'][BEGIN]:
            continue

        if index <= element['ranges']['line'][END]:
            return { 'element': element, 'instruction': element }

        if element['type'] is FIELD:
            match_in_field = check_field_by_index(element, index)
            if match_in_field:
                return match_in_field
        elif element['type'] is FIELDSET:
            match_in_fieldset = check_fieldset_by_index(element, index)
            if match_in_fieldset:
                return match_in_fieldset
        elif element['type'] is LIST:
            match_in_list = check_list_by_index(element, index)
            if match_in_list:
                return match_in_list
        elif element['type'] is MULTILINE_FIELD_BEGIN:
            if not 'template' in element:  # TODO: More elegant copy detection?
                match_in_multiline_field = check_multiline_field_by_index(element, index)
                if match_in_multiline_field:
                    return match_in_multiline_field
        elif element['type'] is SECTION:
            return check_in_section_by_index(element, index)

        break

    return { 'element': section, 'instruction': None }


def lookup(input: str, *, column=None, index=None, line=None, **options):
    context = Context(input, **options)

    match = None
    if index is None:
        if line < 0 or line >= context.line_count:
            raise IndexError(f"You are trying to look up a line ({line}) outside of the document's line range (0-{context.line_count - 1})")

        match = check_in_section_by_line(context.document, line)
    else:
        if index < 0 or index > len(context.input):
            raise IndexError(f"You are trying to look up an index ({index}) outside of the document's index range (0-{len(context.input)})")

        match = check_in_section_by_index(context.document, index)

    result = {
        'element': Element(context, match['element']),
        'range': None
    }

    instruction = match['instruction']

    if not instruction:
        if index is None:
            instruction = next((i for i in context.meta if i['line'] == line), None)
        else:
            instruction = next((i for i in context.meta if index >= i['ranges']['line'][BEGIN] and index <= i['ranges']['line'][END]), None)

        if not instruction:
            return result

    rightmost_match = instruction['ranges']['line'][0]

    if index is None:
        index = instruction['ranges']['line'][0] + column

    for type, range in instruction['ranges'].items():
        if type == 'line':
            continue

        if index >= range[BEGIN] and index <= range[END] and range[BEGIN] >= rightmost_match:
            result['range'] = type
            # TODO: Provide content of range too as convenience
            rightmost_match = index

    return result



from ..errors.validation import Validation
from . import fieldset_entry
from . import section
from .element_base import ElementBase
from .missing import missing_fieldset_entry


class Fieldset(ElementBase):
    def __init__(self, context, instruction, parent=None):
        super().__init__(context, instruction, parent)

        self._all_entries_required = parent._all_elements_required if parent else False

    def __repr__(self):
        return f"<class Fieldset key={self._instruction['key']} entries={len(self._entries())}>"

    def _entries(self, as_map=False):
        if not hasattr(self, '_instantiated_entries'):
            self._instantiated_entries = []
            self._instantiated_entries_map = {}
            self._instantiate_entries(self._instruction)

        return self._instantiated_entries_map if as_map else self._instantiated_entries

    def _entry(self, key, *, required=False):
        self._touched = True

        if not key:
            entries = self._entries()
        else:
            entries_map = self._entries(True)
            entries = entries_map[key] if key in entries_map else []

        if len(entries) == 0:
            if required or self._all_entries_required:
                raise Validation.missing_element(self._context, key, self._instruction, 'missing_fieldset_entry')
            elif required is None:
                return missing_fieldset_entry.MissingFieldsetEntry(key, self)
            else:
                return None

        if len(entries) > 1:
            raise Validation.unexpected_multiple_elements(
                self._context,
                key,
                [e._instruction for e in entries],
                'expected_single_fieldset_entry'
            )

        return entries[0]

    def _instantiate_entries(self, fieldset):
        if 'mirror' in fieldset:
            self._instantiate_entries(fieldset['mirror'])
        elif 'entries' in fieldset:
            def instantiate_and_index(entry):
                instance = fieldset_entry.FieldsetEntry(self._context, entry, self)

                if entry['key'] in self._instantiated_entries_map:
                    self._instantiated_entries_map[entry['key']].append(instance)
                else:
                    self._instantiated_entries_map[entry['key']] = [instance]

                return instance

            filtered = [entry for entry in fieldset['entries'] if entry['key'] not in self._instantiated_entries_map]
            native_entries = [instantiate_and_index(entry) for entry in filtered]

            if 'extend' in fieldset:
                self._instantiate_entries(fieldset['extend'])

            self._instantiated_entries.extend(native_entries)

    def _missingError(self, entry):
        raise Validation.missing_element(self._context, entry._key, self._instruction, 'missing_fieldset_entry')

    def _untouched(self):
        if not hasattr(self, '_touched'):
            return self._instruction

        untouched_entry = next((entry for entry in self._entries() if not hasattr(entry, '_touched')), None)

        return untouched_entry._instruction if untouched_entry else False

    def all_entries_required(self, required=True):
        self._all_entries_required = required

    def assert_all_touched(self, message=None, *, only=None, skip=None):
        entries_map = self._entries(True)

        for key, entries in entries_map.items():
            if (skip and key in skip) or (only and key not in only):
                continue

            for entry in entries:
                if not hasattr(entry, '_touched'):
                    if callable(message):
                        message = message(entry)

                    raise Validation.unexpected_element(self._context, message, entry['instruction'])

    def entries(self, key=None):
        self._touched = True

        if not key:
            return self._entries()
        else:
            entries_map = self._entries(True)
            return entries_map[key] if key in entries_map else []

    def entry(self, key=None):
        return self._entry(key)

    def optional_entry(self, key=None):
        return self._entry(key, required=False)

    def parent(self):
        return self._parent or section.Section(self._context, self._instruction['parent'])

    def required_entry(self, key=None):
        return self._entry(key, required=True)

    def touch(self):
        self._touched = True

        for entry in self._entries():
            entry._touched = True



import math
from .analyzer import Analyzer
from .messages import en
from .reporters.text_reporter import TextReporter
from .resolver import Resolver
from .constants import (
    BEGIN,
    DOCUMENT,
    EMPTY_ELEMENT,
    FIELD,
    FIELDSET,
    FIELDSET_ENTRY,
    LIST,
    LIST_ITEM,
    MULTILINE_FIELD_BEGIN,
    PRETTY_TYPES,
    SECTION
)

class Context:
    def __init__(self, input: str, *, locale=en, reporter=TextReporter, source=None):
        self.input = input
        self.messages = locale
        self.reporter = reporter
        self.source = source

        self.document = {
            'elements': [],
            'type': DOCUMENT
        }

        self.meta = []

        Analyzer(self).analyze()

        if hasattr(self, '_copy_non_section_elements') or hasattr(self, '_copy_sections'):
            Resolver(self).resolve()

    def comment(self, element):
        if not 'computed_comment' in element:
            if 'comments' in element:
                if len(element['comments']) == 1:
                    element['computed_comment'] = element['comments'][0]['comment']
                else:
                    first_non_empty_line_index = None
                    shared_indent = math.inf
                    last_non_empty_line_index = None

                    for index, comment in enumerate(element['comments']):
                        if 'comment' in comment:
                            if first_non_empty_line_index is None:
                                first_non_empty_line_index = index

                            indent = comment['ranges']['comment'][BEGIN] - comment['ranges']['line'][BEGIN]
                            if indent < shared_indent:
                                shared_indent = indent

                            last_non_empty_line_index = index

                    if first_non_empty_line_index is not None:
                        non_empty_lines = element['comments'][first_non_empty_line_index:last_non_empty_line_index + 1]

                        def indent(comment):
                            if not 'comment' in comment:
                                return ''
                            elif (comment['ranges']['comment'][BEGIN] - comment['ranges']['line'][BEGIN]) == shared_indent:
                                return comment['comment']
                            else:
                                return (' ' * (comment['ranges']['comment'][BEGIN] - comment['ranges']['line'][BEGIN] - shared_indent)) + comment['comment']

                        element['computed_comment'] = '\n'.join(indent(comment) for comment in non_empty_lines)
                    else:
                        element['computed_comment'] = None
            else:
                element['computed_comment'] = None

        return element['computed_comment']

    def elements(self, section):
        if 'mirror' in section:
            self.elements(section['mirror'])
        else:
            if not 'computed_elements' in section:
                section['computed_elements_map'] = {}
                section['computed_elements'] = section['elements']

                for element in section['computed_elements']:
                    if element['key'] in section['computed_elements_map']:
                        section['computed_elements_map'][element['key']].append(element)
                    else:
                        section['computed_elements_map'][element['key']] = [element]

                if 'extend' in section:
                    copied_elements = [element for element in self.elements(section['extend']) if element['key'] not in section['computed_elements_map']]

                    section['computed_elements'] = copied_elements + section['computed_elements']

                    for element in copied_elements:
                        if element['key'] in section['computed_elements_map']:
                            section['computed_elements_map'][element['key']].append(element)
                        else:
                            section['computed_elements_map'][element['key']] = [element]

            return section['computed_elements']

    def entries(self, fieldset):
        if 'mirror' in fieldset:
            return self.entries(fieldset['mirror'])

        if not 'computed_entries' in fieldset:
            fieldset['computed_entries_map'] = {}
            fieldset['computed_entries'] = fieldset['entries']

            for entry in fieldset['computed_entries']:
                if entry['key'] in fieldset['computed_entries_map']:
                    fieldset['computed_entries_map'][entry['key']].append(entry)
                else:
                    fieldset['computed_entries_map'][entry['key']] = [entry]

            if 'extend' in fieldset:
                copied_entries = [entry for entry in self.entries(fieldset['extend']) if entry['key'] not in fieldset['computed_entries_map']]

                fieldset['computed_entries'] = copied_entries + fieldset['computed_entries']

                for entry in copied_entries:
                    if entry['key'] in fieldset['computed_entries_map']:
                        fieldset['computed_entries_map'][entry['key']].append(entry)
                    else:
                        fieldset['computed_entries_map'][entry['key']] = [entry]

        return fieldset['computed_entries']

    def items(self, list):
        if 'mirror' in list:
            return self.items(list['mirror'])

        if 'extend' in list:
            if not 'computed_items' in list:
                list['computed_items'] = self.items(list['extend']) + list['items']

            return list['computed_items']

        if 'items' in list:
            return list['items']

        return []

    def raw(self, element):
        result = { 'type': PRETTY_TYPES[element['type']] }

        if 'comments' in element:
            result['comment'] = self.comment(element)

        if element['type'] == EMPTY_ELEMENT:
            result['key'] = element['key']
        elif element['type'] == FIELD:
            result['key'] = element['key']
            result['value'] = self.value(element)
        elif element['type'] == LIST_ITEM:
            result['value'] = self.value(element)
        elif element['type'] == FIELDSET_ENTRY:
            result['key'] = element['key']
            result['value'] = self.value(element)
        elif element['type'] == MULTILINE_FIELD_BEGIN:
            result['key'] = element['key']
            result['value'] = self.value(element)
        elif element['type'] == LIST:
            result['key'] = element['key']
            result['items'] = [self.raw(item) for item in self.items(element)]
        elif element['type'] == FIELDSET:
            result['key'] = element['key']
            result['entries'] = [self.raw(entry) for entry in self.entries(element)]
        elif element['type'] == SECTION:
            result['key'] = element['key']
            result['elements'] = [self.raw(section_element) for section_element in self.elements(element)]
        elif element['type'] == DOCUMENT:
            result['elements'] = [self.raw(section_element) for section_element in self.elements(element)]

        return result

    def value(self, element):
        if 'computed_value' not in element:
            if 'mirror' in element:
                return self.value(element['mirror'])

            element['computed_value'] = None

            if element['type'] is MULTILINE_FIELD_BEGIN:
                if 'lines' in element:
                    element['computed_value'] = self.input[
                        element['lines'][0]['ranges']['line'][0]:element['lines'][-1]['ranges']['line'][1]
                    ]
            else:
                if 'value' in element:
                    element['computed_value'] = element['value']

                if 'continuations' in element:
                    unapplied_spacing = None

                    for continuation in element['continuations']:
                        if element['computed_value'] is None:
                            if 'value' in continuation:
                                element['computed_value'] = continuation['value']

                            unapplied_spacing = None
                        elif not 'value' in continuation:
                            unapplied_spacing = unapplied_spacing or 'spaced' in continuation
                        elif 'spaced' in continuation or unapplied_spacing:
                            element['computed_value'] += ' ' + continuation['value']
                            unapplied_spacing = None
                        else:
                            element['computed_value'] += continuation['value']

        return element['computed_value']



from ..errors.validation import Validation
from . import section
from .value_element_base import ValueElementBase


class Field(ValueElementBase):
    def __repr__(self):
        return f"<class Field key={self._instruction['key']} value={self._print_value()}>"

    def _value(self, loader=None, *, required):
        self._touched = True

        value = self._context.value(self._instruction)

        if value is None:
            if required:
                raise Validation.missing_value(self._context, self._instruction)

            return None

        # TODO: Check if ValueError is 1) a valid pythonic approach 2) the best choice

        if not loader:
            return value

        try:
            return loader(value)
        except ValueError as message:
            raise Validation.value_error(self._context, message, self._instruction)


    def optional_string_value(self):
        return self._value(required=False)

    def optional_value(self, loader):
        return self._value(loader, required=False)

    def parent(self):
        return self._parent or section.Section(self._context, self._instruction.parent)

    def required_string_value(self):
        return self._value(required=True)

    def required_value(self, loader):
        return self._value(loader, required=True)
